import Bio
import sys
import math
from Bio import SeqIO


#declare input
background_files = ["ENCFF002DGZ", "ENCFF002DHJ"]
eE0 = math.exp(2.848039)
PWM = [0.99377,0.002394,0.000393,0.000303,0.999999,0.000407,#
3.7e-05,9e-06,0.0,0.999696,0.0,0.001929,#
0.0,0.0,0.999607,0.0,0.0,3e-06,#
0.006192,0.997597,0.0,0.0,1e-06,0.997661]
L_w = len(PWM) // 4


#define nucleic acid dictionary
nucl_dict = {"A" : 0, "C" : 1, "G" : 2, "T" : 3}
inv_nucl_dict = {0 : "A", 1 : "C", 2 : "G", 3 : "T"}

#save original output channel
original_stdout = sys.stdout


#create array of background counts
background_frequency = [0.0 for i in range(0,pow(4,L_w))]


#define function which calculates the background frequency of a given read S
def construct_frequency(S):
	ret = 0.0
	
	for kmer_start in range(0,len(S)-L_w+1):
		tmp_pos = 0
		
		for kmer_rel_pos in range(0,L_w):
			tmp_pos += nucl_dict[ S[kmer_start+kmer_rel_pos] ] * pow(4,kmer_rel_pos)
		
		ret += background_frequency[tmp_pos]
	
	return ret


#function determines exp(E(S))
#position-specific binding likelihood
def eES(read):
	ret = 0.0
	for kmer_start in range(0,len(read) - L_w + 1):
		aux = 1.0
		for kmer_rel_pos in range(0,L_w):
			aux *= PWM[nucl_dict[read[kmer_start+kmer_rel_pos]]*L_w + kmer_rel_pos]
		if 'N' not in read[kmer_start:kmer_start+L_w]:
			ret += aux
	return ret




#-------------------------------
#START OF THE SCRIPT
#-------------------------------
#construct background frequency array
#normalization = 0
#for data_file in background_files:
	#for read in SeqIO.parse(data_file+".fastq","fastq"):
		#s = read.seq
		
		#for kmer_start in range(0,len(s)-L_w+1):
			#tmp_cnt = 0
			#is_N = 0
			
			#for kmer_rel_pos in range(0,L_w):
				#if(s[kmer_start+kmer_rel_pos] != 'N'):
					#tmp_cnt += nucl_dict[ s[kmer_start + kmer_rel_pos] ] * pow(4,kmer_rel_pos)
				#else:
					#is_N = 1
					
			##only use sequence if there is no N in it
			#if( is_N == 0 ):
				#background_frequency[tmp_cnt] += 1.0
				#normalization += 1
			
		
	

##normalize
#for i in range(0,pow(4,L_w)):
	#background_frequency[i] = background_frequency[i]/normalization


##create background sequence dictionaries, print in data format for RBAPBindnSeq
#number_kmers = 0
#for data_file in background_files:
	##print read files with background frequencies
	#with open(data_file+".dat","w") as f:
		#sys.stdout = f
		#for seq_record in SeqIO.parse(data_file+".fastq","fastq"):
			#if('N' not in seq_record.seq AND len(seq_record.seq) >= L_w):
				#print(len(seq_record.seq),1,construct_frequency(seq_record.seq),seq_record.seq)
		#sys.stdout = original_stdout
		
	
bgr_log_likelihood = 0.0
norm1 = 0
norm2 = 0.0
for data_file in background_files:
	f = open(data_file+'.dat','r')
	
	while True:
		line = f.readline()
		
		if not (line):
			break
		
		val = line.split()
		
		peES = 0.0
		if len(val) == 4:
			peES = eES(val[3])
			peES *= float(val[2])
			peES += eE0*(int(val[0]) - L_w + 1)
		
		if peES > 0.0:
			bgr_log_likelihood += int(val[1])*math.log(peES)
			norm1 += int(val[1])
			norm2 += peES
		
	

print("Background normalization",bgr_log_likelihood-norm1*math.log(norm2))

