#script searches output file for PWMs
#subtracts the respective background probabilities
#prints a ranked list of accordingly ranked PWMs

import math
import os
import sys

#specify prefix for files
prefix = "data"

#define thresholds
spec_threshhold = 50.0

#give background values
normalization_dict = {4 : 6044491827, 5 : 5880088810, 6 : 5715685929, 7 : 5551283214}

#save original output channel
original_stdout = sys.stdout

#define nucleic acid dictionary
nucl_dict = {"A" : 0, "C" : 1, "G" : 2, "T" : 3 }
inv_nucl_dict = {0 : "A", 1 : "C", 2 : "G", 3 : "T" }



#define class of PWMs with score
class PWM_Score():
	def __init__(self, PWM, foreground_log_likelihood, iterations, E0):
		self.PWM = PWM
		self.foreground_log_likelihood = foreground_log_likelihood
		self.iterations = iterations
		self.E0 = E0


#--------------------
#START SCRIPT
#--------------------

#initialize PWM collection
PWM_score_array = []

for fi in os.listdir(os.getcwd()):
	filename = os.fsdecode(fi)
	
	if ((not filename.startswith("ENCFF")) and filename.endswith("mer.dat") and filename.startswith(prefix)): 
		
		foreground_file = filename
		
		#open files
		f = open(foreground_file,'r')
		
		while True:
			# Get next PWM from file
			line1 = f.readline()	#skip first line
			line1 = f.readline()
			line2 = f.readline()
			line3 = f.readline()
			line4 = f.readline()
			line5 = f.readline()
			
			# if two lines are empty    
			# end of file is reached
			if not (line1):
				break
			 
			#extract words from lines
			val1 = line1.split()
			val2 = line2.split()
			val3 = line3.split()
			val4 = line4.split()
			val5 = line5.split()
			
			#skip free line after each PWM
			line1 = f.readline()
			
			Lstart = float(val5[1])
			
			line1 = f.readline()	#skip first line
			line1 = f.readline()
			line2 = f.readline()
			line3 = f.readline()
			line4 = f.readline()
			line5 = f.readline()
			
			# if two lines are empty    
			# end of file is reached
			if not (line1):
				break
			 
			#extract words from lines
			val1 = line1.split()
			val2 = line2.split()
			val3 = line3.split()
			val4 = line4.split()
			val5 = line5.split()
			
			#skip free line after each PWM
			line1 = f.readline()
			
			Lstop = float(val5[1])
			
			if (Lstart < Lstop and float(val5[0]) < spec_threshhold):
				Lw = len(val1) - 1
				pPWM = [0.0 for j in range(0,4*Lw)]
				for i in range(0,Lw):
					pPWM[0*Lw+i] = float(val1[i+1])
				for i in range(0,Lw):
					pPWM[1*Lw+i] = float(val2[i+1])
				for i in range(0,Lw):
					pPWM[2*Lw+i] = float(val3[i+1])
				for i in range(0,Lw):
					pPWM[3*Lw+i] = float(val4[i+1])
				tmp_foreground = Lstop #/ normalization_dict[Lw]
				tmp_iterations = int(float(val5[2]))
				tmp_E0 = float(val5[0])
				tmp_obj = PWM_Score(PWM = pPWM, foreground_log_likelihood = tmp_foreground, iterations = tmp_iterations, E0 = tmp_E0)
				PWM_score_array.append(tmp_obj)
			
			
		
		f.close()
		
#with open(filename[:-4]+"_ranked.dat","w") as f:
with open("PWMs_ranked.dat","w") as f:
	sys.stdout = f
	#print ranked consensus PWMs
	for p in sorted(PWM_score_array, key= lambda item: item.foreground_log_likelihood, reverse= True):
		print("", end = "\t")
		for i in range(0,int(len(p.PWM)/4)):
			print((i+1), end = "\t\t")
		print("")
		for i in range(0,4):
			print(inv_nucl_dict[i], end = "\t")
			for j  in range(0,int(len(p.PWM)/4)):
				print(p.PWM[i*int(len(p.PWM)/4)+j], end = "\t")
			print("")
		print(p.E0, p.foreground_log_likelihood, p.iterations)
		print("")
	sys.stdout = original_stdout
